# -*- coding: utf-8 -*-
#
# This file is part of Chronoscope.
#
# SPDX-FileCopyrightText: 2024 Anatoliy Bilenko <anatoliy.bilenko@gmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

import chronoscope.db as db
import chronoscope.utils as utils
import matplotlib.cm as cm              # type: ignore
import matplotlib.pyplot as pt          # type: ignore
import matplotlib.ticker as ticker      # type: ignore
from dataclasses import dataclass
import sys


Y_LINE_SPACING = 2
X_TICKS_MAX = 5

def plot_timeline(timeline, y_pos: int):
    y_pos_scaled = -Y_LINE_SPACING * y_pos
    colors = cm.get_cmap("tab10").colors  # type: ignore[attr-defined]

    for current, current_tick in enumerate(timeline[:-1]):
        next_tick = timeline[current + 1]
        start_time, end_time = current_tick["time"], next_tick["time"]
        event_label = current_tick["event"]

        pt.hlines(y_pos_scaled, start_time, end_time, lw=4,
                  colors=colors[current % len(colors)])
        pt.text(start_time, y_pos_scaled, event_label, rotation=90)

    pt.text(timeline[-1]["time"], y_pos_scaled,
            timeline[-1]["event"], rotation=90)

@dataclass
class timeline_visitor:
    y_labels: list
    y_pos: int
    x_min: int
    x_max: int

    def __call__(self, timeline: list[dict], origin: int, parent: None | int):
        plot_timeline(timeline, self.y_pos)
        self.y_pos += 1
        duration = round((timeline[-1]["time"] - timeline[0]["time"]) / 1e6, 3)
        type, id = timeline[0]["type"], timeline[0]["id"]
        self.y_labels.append(f"{type}{utils.unpack(id)} [{duration}ms]")

        times = [tick["time"] for tick in timeline]
        self.x_min = min(self.x_min, min(times))
        self.x_max = max(self.x_max, max(times))

class chart_annotation:
    def __init__(self, fig):
        self.cur_mark = ord('A')
        self.cur = []
        self.ann = []
        self.ann_mode = False
        fig.canvas.mpl_connect("key_press_event", self.on_key)
        fig.canvas.mpl_connect("button_press_event", self.on_click)

    def on_key(self, event):
        if event.key == "escape":
            sys.exit(1)
        elif event.key == "e":
            self.ann_mode = not self.ann_mode
        elif event.key == "d":
            if self.ann:
                self.cur_mark -= 1
                self.cur.pop()
                self.ann.pop().remove()
                self.ann.pop().remove()
            event.canvas.draw()

    def on_click(self, event):
        if not self.ann_mode:
            return

        x, ax = event.xdata, event.inaxes
        # TODO: Don't ask me what's going on with cursors... I don't know!
        self.cur.append(x)
        cursor = chr(self.cur_mark)
        if self.cur_mark % 2 == 0:
            cursor = chr(self.cur_mark - 1) + cursor + ": "
            cursor += utils.str_ns_diff(int(abs(self.cur[-1] - self.cur[-2])))
        self.cur_mark += 1

        self.ann.append(ax.axvline(x=x, color="lightgray"))
        self.ann.append(ax.annotate(cursor, xycoords=("data", "axes fraction"),
                                    xy=(x, 0)))
        self.ann_mode = not self.ann_mode
        event.canvas.draw()

def plot(origin: int, figsize=(16, 4), depth_max=50):
    fig = pt.figure(figsize=figsize)
    pt.style.use("bmh")
    pt.rcParams["font.size"] = 8
    pt.subplots_adjust(top=0.75)

    v = timeline_visitor([], 0, utils.MAX_INT, utils.MIN_INT)
    db.iterate(origin, None, db.tick, v, 0, depth_max)

    end = -Y_LINE_SPACING * v.y_pos
    y_range = [float(x) for x in range(0, end, -Y_LINE_SPACING)]
    x_range = range(v.x_min, v.x_max, round((v.x_max - v.x_min) / X_TICKS_MAX))

    # (1)
    # x_labels = [utils.str_ns(x, compact=True) for x in x_range]
    # pt.xticks(x_range, x_labels)
    # (2) This might be slow, consider to uncomment (1) in such cases.
    pt.gca().xaxis.set_major_formatter(ticker.FuncFormatter(
        lambda value, pos: utils.str_ns(value, compact=True)))

    pt.yticks(y_range, v.y_labels)
    pt.xlabel("Time")
    pt.autoscale(enable=True, axis="x", tight=True)
    pt.margins(0.1)
    _ = chart_annotation(fig)

    pt.grid(True)
    title = f"Request {utils.unpack(origin)}\n"
    title += f"[{utils.str_ns(x_range[0])}...{utils.str_ns(x_range[-1])}]"
    pt.suptitle(title)
    pt.show()



# -*- coding: utf-8 -*-
#
# This file is part of Chronoscope.
#
# SPDX-FileCopyrightText: 2024 Anatoliy Bilenko <anatoliy.bilenko@gmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

import chronoscope.parser as pr
from typing import Callable
import subprocess as sp
import builtins as b
import peewee as p
import os

db = p.SqliteDatabase(None)

class T(p.Model):
    class Meta:
        database = db
        primary_key = False

class tick(T):
    id = p.IntegerField()
    time = p.IntegerField()
    event = p.TextField()
    type = p.TextField()

class attr(T):
    id = p.IntegerField()
    name = p.TextField()
    val = p.TextField()

class relation(T):
    orig = p.IntegerField()
    dest = p.IntegerField()
    type = p.TextField()


TABLES = [tick, attr, relation]

def open(path: str, opts: None | dict[str, int | str] = None, create=False):
    if create and os.path.exists(path):
        raise FileExistsError("`{path}' exists!")
    if not create and not os.path.exists(path):
        raise FileNotFoundError("`{path}' not found!")

    db.init(path, opts)
    db.connect()
    if create:
        with db:
            db.create_tables(TABLES)

def close():
    db.close()

def mkidx():
    db.execute_sql("CREATE INDEX tick_idx on tick(id);")
    db.execute_sql("CREATE INDEX relation_idx on relation(orig);")
    db.execute_sql("CREATE INDEX attr_idx on attr(id);")

def line_nr(file: str) -> int:
    result = sp.run(['wc', file], stdout=sp.PIPE, text=True)
    return int(result.stdout.split()[0])

def load(pr: pr.parser, trace_path: str, fd_chunk_size=100, db_chunk_size=10):
    if not os.path.exists(trace_path):
        raise FileNotFoundError("`{trace_path}' not found!")

    with b.open(trace_path) as fd:
        for fd_chunk in p.chunked(fd, fd_chunk_size):
            records = pr.parse(fd_chunk)
            with db.atomic():
                for table in TABLES:
                    t_name: str = table._meta.name  # type: ignore
                    for db_chunk in p.chunked(records[t_name], db_chunk_size):
                        table.insert_many(db_chunk).execute()

def iterate(origin: int, parent: None | int, samples: type[tick] | type[attr],
            visit: Callable, depth: int, depth_max: int):
    if depth_max < depth:
        return

    # pull origin
    if timeline := samples.select().where((samples.id == origin)):
        visit(timeline.dicts(), origin, parent)

    # pull children
    orig_to_children = relation.select().where((relation.orig == origin))
    for child in orig_to_children.dicts():
        print(f"@[{depth}] {hex(child['orig'])} ... {hex(child['dest'])} ...")
        iterate(child["dest"], origin, samples, visit, depth + 1, depth_max)



# -*- coding: utf-8 -*-
#
# This file is part of Chronoscope.
#
# SPDX-FileCopyrightText: 2024 Anatoliy Bilenko <anatoliy.bilenko@gmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

import chronoscope.db as db
import chronoscope.utils as utils
from graphviz import Graph  # type: ignore

# https://graphviz.readthedocs.io/en/stable/examples.html#structs-py
FMT_NODE = """<
<table border="0" cellborder="1" cellspacing="0">
<tr><td>{}</td></tr><tr><td>{}</td></tr></table>>
"""

class attr_visitor:
    def __init__(self, graph: Graph):
        self.graph = graph

    def __call__(self, node_attrs: list[dict], current: int, parent: None | int):
        contents = "<br/>".join([f"{na['name']}={na['val']}" for na in node_attrs])
        self.graph.node(str(current), FMT_NODE.format(utils.unpack(current), contents))

        if parent:
            self.graph.edge(str(parent), str(current))

def plot(origin: int, depth_max=50):
    g = Graph(strict=True, format="png", node_attr={"shape": "plaintext"})
    db.iterate(origin, None, db.attr, attr_visitor(g), 0, depth_max)
    pid, id = utils.unpack(origin)
    g.render(f"tree_{pid}_{id}", cleanup=True)



# -*- coding: utf-8 -*-
#
# This file is part of Chronoscope.
#
# SPDX-FileCopyrightText: 2024 Anatoliy Bilenko <anatoliy.bilenko@gmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

from chronoscope import db
from chronoscope import parser
import chronoscope.tree as tree
import chronoscope.chart as chart
import argparse
import sys

__version__ = '0.0.1'
__author__ = 'Anatoliy Bilenko <anatoliy.bilenko@gmail.com>'
__license__ = 'LGPLv3'

db_options: dict[str, int | str] = {
    "cache_size": -128 << 20,
    "synchronous": "off",
    "journal_mode": "off",
}

def parse_args():
    parser = argparse.ArgumentParser(prog=sys.argv[0], description="""
    chronoscope: Plots drill-down chart based on user's traces""")
    parser.add_argument("-d", "--db", type=str, default="chronoscope.db",
                        help="chronoscope database")
    parser.add_argument("-c", "--conf", type=str, default="chronoscope.yaml",
                        help="configuration, defines how to parse out users'\n"
                        "ticks, attrs and relations from the traces")
    parser.add_argument("-t", "--trace", type=str, help="User's traces")
    parser.add_argument("-D", "--depth", type=int, default=50,
                        help="limits output to given level of ticks")
    parser.add_argument("command", type=str, choices=["create", "plot"],
                        help="create: build chronoscope database\n"
                        "plot: plots drill-down chart for the given tick")
    parser.add_argument("-f", "--fig_size", nargs=2, type=int, default=[16, 4])
    parser.add_argument("-k", "--tick_id", type=int,
                        help="tick identifier to plot")
    return parser.parse_args()


def main() -> int:
    args = parse_args()

    if args.command == "create":
        db.open(args.db, db_options, create=True)
        db.load(parser.parser(args.conf), args.trace)
        db.mkidx()
        db.close()

    if args.command == "plot":
        db.open(args.db, db_options)
        chart.plot(args.tick_id, args.fig_size, args.depth)
        tree.plot(args.tick_id, args.depth)

    return 0



# -*- coding: utf-8 -*-
#
# This file is part of Chronoscope.
#
# SPDX-FileCopyrightText: 2024 Anatoliy Bilenko <anatoliy.bilenko@gmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

from datetime import datetime as t
import sys

NS_TIME_LEN = len("2055-11-29T20:57:56.489282133")
FMT_MS = "%Y-%m-%dT%H:%M:%S.%f"
FMT_MS_COMPACT = ":%S.%f"
MAX_INT = sys.maxsize
MIN_INT = -sys.maxsize - 1
BITS_PER_PID = 8
DB_INT_SIZE = 64


def pack_unpack_init(p_bits_per_pid: int, p_db_int_size: int):
    global BITS_PER_PID, DB_INT_SIZE
    BITS_PER_PID = p_bits_per_pid
    DB_INT_SIZE = p_db_int_size

def pack(id: int, pid: int) -> int:
    masked_pid = ((1 << BITS_PER_PID) - 1) & pid
    packed_pid = masked_pid << (DB_INT_SIZE - BITS_PER_PID)
    packed_id = ((1 << (DB_INT_SIZE - BITS_PER_PID)) - 1) & id
    return (packed_pid | packed_id)

def unpack(id_pid: int) -> tuple[int, int]:
    pid = id_pid >> (DB_INT_SIZE - BITS_PER_PID)
    id = ((1 << (DB_INT_SIZE - BITS_PER_PID)) - 1) & id_pid
    return pid, id

def ns(time: str) -> int:
    if len(time) != NS_TIME_LEN:
        raise ValueError("Not a nanosecond time format")
    ms = int(t.strptime(time[:-3], FMT_MS).timestamp() * 1e6)
    return ms * 1_000 + int(time[-3:])

def str_ns(unix_time_ns: int, compact=False) -> str:
    dt = t.utcfromtimestamp(unix_time_ns / 1e9)
    if compact:
        return dt.strftime(FMT_MS_COMPACT)
    return dt.strftime(FMT_MS)

def str_ns_diff(unix_time_ns: int) -> str:
    return str(unix_time_ns) + "ns"



# -*- coding: utf-8 -*-
#
# This file is part of Chronoscope.
#
# SPDX-FileCopyrightText: 2024 Anatoliy Bilenko <anatoliy.bilenko@gmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

import chronoscope
import sys

if __name__ == "__main__":
    sys.exit(chronoscope.main())



import os
import re
import sys
from io import open
from setuptools import setup


if sys.version_info < (3, 10):
    print('Chronoscope requires at least Python 3.10 to run.')
    sys.exit(1)

with open(os.path.join('chronoscope', '__init__.py'), encoding='utf-8') as f:
    rexp = r"^__version__ = ['\"]([^'\"]*)['\"]"
    version = re.search(rexp, f.read(), re.M).group(1)  # type: ignore


if not version:
    raise RuntimeError('Cannot find Chronoscope version information.')

def install_requires():
    requires = [
        'peewee',
        'matplotlib',
        'PyYAML',
        'graphviz',
        'packaging',
    ]

    return requires


setup(
    name='Chronoscope',
    version=version,
    description="A cross-platform matplotlib-based observability tool",
    long_description="A cross-platform matplotlib-based observability tool",
    author='Anatoliy Bilenko',
    author_email='anatoliy.bilenko@gmail.com',
    url='https://github.com/just-now/chronoscope',
    license='LGPLv3',
    keywords="cli observability",
    python_requires=">=3.10",
    install_requires=install_requires(),
    packages=['chronoscope'],
    entry_points={"console_scripts": ["chronoscope=chronoscope:main"]},
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Topic :: System :: Monitoring'
    ]
)



# -*- coding: utf-8 -*-
#
# This file is part of Chronoscope.
#
# SPDX-FileCopyrightText: 2024 Anatoliy Bilenko <anatoliy.bilenko@gmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

import chronoscope.utils as u
from typing import Callable
import yaml

class parser:
    def __init__(self, conf_path: str):
        # type -> (dest_table, parser_func)
        self.parsers: dict[str, tuple[str, Callable]] = {}
        # the parser knows about table names
        self.tables = ["tick", "attr", "relation"]
        self.load_config(conf_path)

    def load_config(self, conf_path: str):
        with open(conf_path) as fd:
            conf = yaml.safe_load(fd)
            conf_tables = [t for t in conf.keys() if t[0] != '.']
            if not all(t in self.tables for t in conf_tables):
                raise SyntaxError(f"a few of {conf_tables} aren't known!")

            for table in conf_tables:
                for trace in conf[table]:
                    self.register_parser(table, trace["type"],
                                         self.make_parser(table, trace["pos"]))

    def make_parser(self, table: str, kwargs: dict[str, int]) -> Callable:
        match table:
            case "tick":
                return self.make_req_parser(**kwargs)
            case "relation":
                return self.make_rel_parser(**kwargs)
            case "attr":
                return self.make_attr_parser(**kwargs)
        raise NotImplementedError()

    def make_req_parser(self, type: int, time: int, event: int,
                        pid: int, id: int) -> Callable:
        def parse(line: list[str], parse_type: str):
            return {
                "time": u.ns(line[time]),
                "type": line[type], "event": line[event],
                "id": u.pack(int(line[id]), int(line[pid]))
            } if parse_type == line[type] else None
        return parse

    def make_rel_parser(self, orig_id: int, dest_id: int,
                        orig_pid: int, dest_pid: int, type: int) -> Callable:
        def parse(line: list[str], parse_type: str):
            return {
                "orig": u.pack(int(line[orig_id]), int(line[orig_pid])),
                "dest": u.pack(int(line[dest_id]), int(line[orig_pid])),
                "type": line[type]
            } if parse_type == line[type] else None
        return parse

    def make_attr_parser(self, id: int, pid: int,
                         name: int, value: int, type: int) -> Callable:
        def parse(line: list[str], parse_type: str):
            return {
                "id": u.pack(int(line[id]), int(line[pid])),
                "val": line[value],
                "name": line[name],
            } if parse_type == line[type] else None
        return parse

    def register_parser(self, dest_table: str, type: str, parse: Callable):
        self.parsers[type] = (dest_table, parse)

    def parse(self,
              fd_chunk: list[str]) -> dict[str, list[dict[str, str | int]]]:
        records: dict[str, list] = {t: [] for t in self.tables}
        for line in fd_chunk:
            for p_name, (dest_table, parser) in self.parsers.items():
                if record := parser(line.split(), p_name):
                    records[dest_table].append(record)
        return records
